/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


import java.applet.Applet;
import java.awt.Graphics;
import java.awt.Component;
import java.awt.Container;
import java.awt.BorderLayout;
import java.net.URL;
import java.net.MalformedURLException;


public class vmApplet extends Applet implements Runnable, vmConst {
  final static String VERSION = "vmApplet 1.0.42c";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  private vmEngine vm;
  private vmVideo vmOut;
  private vmTerm vmIn;
  private vmSystem vmSys;
  private Thread appletThread = null;
  private String byeURL;
  public vmApplet() {}

  public void init() {
    int MEMORY_SIZE = 0x00007FFF;
    int CODE_ORG = 0;
    int TIB_SIZE = 80;
    int STACK_SIZE = 128;
    int RETURN_SIZE = 128;
    vmLoader vmLoad;
    String param;
    int i = 0;
    int rows = 25;
    int cols = 72;
    int fontsize = 12;
    int foreground = 0x00FF00;
    int background = 0x000000;

    try {
      param = getParameter("bye");
      if (param == null) {
        param = "http://www.geocities.com/dolphinconsultant/homepage.html";
      }
      byeURL = param;
      param = getParameter("rows");
      if (param != null) {
        rows = getInteger(param);
      }
      param = getParameter("cols");
      if (param != null) {
        cols = getInteger(param);
      }
      param = getParameter("fontsize");
      if (param != null) {
        fontsize = getInteger(param);
      }
      param = getParameter("foreground");
      if (param != null) {
        foreground = getInteger(param);
      }
      param = getParameter("background");
      if (param != null) {
        background = getInteger(param);
      }
      vmOut = new vmVideo(rows, cols, fontsize, foreground, background);
      vmOut.setSize(this.getSize());
      this.setLayout(new BorderLayout());
      this.add(vmOut);
      vmOut.clearPage();
      vmIn = new vmTerm(vmOut);
      param = getParameter("log");
      if (param != null) {
        Integer val = new Integer(param);

        vmOut.setLogEnable(val.intValue());
      }
      int hashSize = 0;

      param = getParameter("hash-init");
      if (param != null) {
        hashSize = getInteger(param);
      }
      vmSys =
          new vmSystem(this, vmIn, vmOut, hashSize,
            getParameter("timezone"));
      param = getParameter("mem-size");
      if (param != null) {
        MEMORY_SIZE = getInteger(param);
      }
      vm = new vmEngine(vmSys, (short) MEMORY_SIZE);
      vmSys.vm = vm;
      param = getParameter("code-org");
      if (param != null) {
        CODE_ORG = getInteger(param);
      }
      param = getParameter("tib-size");
      if (param != null) {
        TIB_SIZE = getInteger(param);
      }
      param = getParameter("stack-size");
      if (param != null) {
        STACK_SIZE = getInteger(param);
      }
      param = getParameter("return-size");
      if (param != null) {
        RETURN_SIZE = getInteger(param);
      }
      vmLoad = new vmLoader(vmSys, vm, MEMORY_SIZE, CODE_ORG,
            TIB_SIZE, STACK_SIZE, RETURN_SIZE);
      vmLoad.prepare();
      param = getParameter("boot");
      if (param != null) {
        for (i = 0; i < param.length(); i++)
          vmIn.enqueueKeyTypedEvent(param.charAt(i));
        vmIn.enqueueKeyTypedEvent((char) 0x0A);
      }
    } catch (Exception e) {
      vm.stop();
      showStatus("Initialization Exception: " + e.toString());
      return;
    }
  }

  public int getInteger(String s0) {
    String s1 = s0.toLowerCase();

    if (s1.charAt(0) == '0' && s1.charAt(1) == 'x') {
      return Integer.parseInt(s1.substring(2, s1.length()), 16);
    } else {
      return Integer.parseInt(s1, 10);
    }
  }

  public void start() {
    if (vm == null)
      init();
    if (!vmOut.isVisible())
      vmOut.setVisible(true);
    if (!vmOut.isEnabled())
      vmOut.setEnabled(true);
    if (!isVisible())
      setVisible(true);
    if (!isEnabled())
      setEnabled(true);
    vmOut.requestFocus();
    if (appletThread == null) {
      appletThread = new Thread(this, "vmApplet");
      appletThread.start();
    }
  }

  public void stop() {
    if (vm != null)
      sysKill();
  }

  public void run() {
    boolean vmInterrupted = false;

    vm.setPriority(Thread.MIN_PRIORITY);
    vm.start();
    try {
      Thread.sleep(500);
    } catch (InterruptedException e) {}
    try {
      vm.join();
    } catch (InterruptedException e) {
      vmInterrupted = true;
    }
    if (!vmInterrupted) {
      vmOut.print(" Goodbye...");
      sysSTOP();
    } else
      sysKill();
  }

  public void sysKill() {
    vm = null;
    appletThread = null;
    setEnabled(false);
    setVisible(false);
    if (vmOut != null) {
      vmOut.setEnabled(false);
      vmOut.setVisible(false);
      vmOut.removeNotify();
      vmIn.close();
      remove(vmOut);
      vmOut = null;
    }
  }

  public void sysSTOP() {
    sysKill();
    URL anchor;

    try {
      anchor = new URL(byeURL);
    } catch (MalformedURLException e) {
      anchor = null;
    }
    if (anchor != null)
      getAppletContext().showDocument(anchor, "_top");
  }

  public void update(Graphics g) {
    paint(g);
  }

  public void paint(Graphics g) {
    if (vmOut != null)
      vmOut.paintScreen();
  }

  public String[][] getParameterInfo() {
    String[][] info = {
        {
          "rows", "int", "applet row count <25>"}, {
          "cols", "int", "applet column count <72>"}, {
          "fontsize", "int", "Courier font size <12>"}, {
          "foreground", "int", "startup foreground color <0x00FF00>"}, {
          "background", "int", "startup background color <0x000000>"}, {
          "bye", "string", "applet exit URL <my homepage>"}, {
          "boot", "string", "startup boot string <null>"}, {
          "log", "int", "screen log enable: 1=TRUE <0>"}, {
          "timezone", "string", "timezone, e.g. GMT <null>"}, {
          "mem-size", "string", "memory size <0x0FFFF>"}, {
          "code-org", "string", "dictionary origin ( UP) <0>"}, {
          "tib-size", "string", "text input buffer size <80>"}, {
          "stack-size", "string", "parameter stack size <128>"}, {
          "return-size", "string", "return stack size <128>"}, {
          "hash-init", "int", "initial hash table size <0>"}, };

    return info;
  }

  public String getAppletInfo() {
    return (_VERSION + " - a forth simulator\n" +
        _COPYRIGHT + "\n"
        +
        "http://www.geocities.com/dolphinconsultant/homepage.html\n" +
        "email: dolphinconsultant@yahoo.com\n");
  }
}
