/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


import java.awt.*;
import java.awt.Component;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Properties;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;


public class vmMain extends Frame implements FocusListener, WindowListener,
    vmConst {
  final static String VERSION = "vmMain 1.0.38";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  private vmEngine vm;
  private vmVideo vmOut;
  private vmTerm vmIn;
  private vmSystem vmSys;
  private Properties properties;
  private boolean inPaint = false;
  public vmMain() {
    super(" myForth");
  }

  private void setup(String args[]) {
    Panel pQuery;
    int fWidth;
    int fHeight;
    StringBuffer sb = new StringBuffer();
    int i = 0;
    int rows = 25;
    int cols = 72;
    int fontsize = 14;
    int foreground = 0x00FF00;
    int background = 0x000000;
    int MEMORY_SIZE = 0x00007FFF;
    int CODE_ORG = 0;
    int TIB_SIZE = 80;
    int STACK_SIZE = 128;
    int RETURN_SIZE = 128;
    int hashSize = 0;
    vmLoader vmLoad;
    String timezone = null;
    String param;
    String configFile = "myforth.cfg";
    String sep = File.separator;
    int iArg = 0;
    String command = null;

    for (; iArg < args.length; iArg++) {
      if (args[iArg].equals("-f")) {
        configFile = args[++iArg];
      }
    }
    if (configFile.indexOf(sep) < 0)
      configFile = System.getProperty("user.dir") + sep + configFile;
    loadProperties(configFile);
    param = properties.getProperty("timezone");
    if (param != null) {
      timezone = param;
    }
    param = properties.getProperty("rows");
    if (param != null) {
      rows = getInteger(param);
    }
    param = properties.getProperty("cols");
    if (param != null) {
      cols = getInteger(param);
    }
    param = properties.getProperty("fontsize");
    if (param != null) {
      fontsize = getInteger(param);
    }
    param = properties.getProperty("foreground");
    if (param != null) {
      foreground = getInteger(param);
    }
    param = properties.getProperty("background");
    if (param != null) {
      background = getInteger(param);
    }
    param = properties.getProperty("hash-init");
    if (param != null) {
      hashSize = getInteger(param);
    }
    param = properties.getProperty("mem-size");
    if (param != null) {
      MEMORY_SIZE = getInteger(param);
    }
    param = properties.getProperty("code-org");
    if (param != null) {
      CODE_ORG = getInteger(param);
    }
    param = properties.getProperty("tib-size");
    if (param != null) {
      TIB_SIZE = getInteger(param);
    }
    param = properties.getProperty("stack-size");
    if (param != null) {
      STACK_SIZE = getInteger(param);
    }
    param = properties.getProperty("return-size");
    if (param != null) {
      RETURN_SIZE = getInteger(param);
    }
    param = properties.getProperty("boot");
    if (param != null) {
      command = "<true>";
      for (i = 0; i < param.length(); i++)
        sb.append(param.charAt(i));
      sb.append(" ");
    }
    if (command != null) {
      sb.append(" ");
      command = new String(sb);
    }
    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();

    this.setLayout(new BorderLayout());
    vmOut = new vmVideo(rows, cols, fontsize, foreground, background);
    vmOut.setSize(d.width, d.height);
    this.add("Center", vmOut);
    this.pack();
    param = properties.getProperty("log");
    if (param != null) {
      Integer val = new Integer(param);

      vmOut.setLogEnable(val.intValue());
    }
    fWidth = vmOut.rowWidth +
        (vmOut.leftPad * 2) + 2 + this.getInsets().left;
    fHeight = (vmOut.topPad * 2) +
        (vmOut.rows * vmOut.height1) + 2 + this.getInsets().top;
    if (fWidth > d.width)
      fWidth = d.width;
    if (fHeight > d.height)
      fHeight = d.height;
    this.setSize(fWidth, fHeight);
    vmOut.setVisible(true);
    vmOut.setEnabled(true);
    setVisible(true);
    setEnabled(true);
    vmOut.clearPage();
    vmIn = new vmTerm(vmOut);
    try {
      vmSys = new vmSystem(this, vmIn, vmOut, hashSize, timezone);
      vm = new vmEngine(vmSys, (short) MEMORY_SIZE);
      vmSys.vm = vm;
      vmLoad = new vmLoader(vmSys, vm, MEMORY_SIZE, CODE_ORG,
            TIB_SIZE, STACK_SIZE, RETURN_SIZE);
      vmLoad.prepare();
      if (command != null) {
        for (i = 0; i < command.length(); i++)
          vmIn.enqueueKeyTypedEvent(command.charAt(i));
        vmIn.enqueueKeyTypedEvent((char) 0x0A);
      }
      vm.setPriority(Thread.MIN_PRIORITY);
      vm.start();
      vmOut.addFocusListener(this);
      addWindowListener(this);
      vmOut.requestFocus();
      vm.join();
      dispose();
      System.exit(0);
    } catch (Exception e) {
      vm.stop();
      System.err.println("Thread Exception: " + e.toString());
      return;
    }
  }

  public static void main(String args[]) {
    vmMain m = new vmMain();

    m.setup(args);
  }

  private void loadProperties(String s0) {
    File propsFile = new File(s0);

    properties = new Properties();
    try {
      FileInputStream in = new FileInputStream(propsFile);

      properties.load(new BufferedInputStream(in));
      in.close();
    } catch (Exception e) {}
  }

  private int getInteger(String s0) {
    String s1 = s0.toLowerCase();

    if (s1.charAt(0) == '0' && s1.charAt(1) == 'x') {
      return Integer.parseInt(s1.substring(2, s1.length()), 16);
    } else {
      return Integer.parseInt(s1, 10);
    }
  }

  public void focusGained(FocusEvent e) {
    if (!inPaint)
      this.repaint();
  }

  public void focusLost(FocusEvent e) {}

  public void windowOpened(WindowEvent e) {};
  public void windowClosing(WindowEvent event) {
    dispose();
    System.exit(0);
  }

  public void windowClosed(WindowEvent e) {};
  public void windowIconified(WindowEvent e) {};
  public void windowDeiconified(WindowEvent e) {};
  public void windowActivated(WindowEvent e) {};
  public void windowDeactivated(WindowEvent e) {};
  public void update(Graphics g) {
    paint(g);
  }

  public void paint(Graphics g) {
    inPaint = true;
    vmOut.paintScreen();
    vmOut.requestFocus();
    inPaint = false;
  }
}
