/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


import java.io.InputStream;
import java.io.IOException;
import java.awt.*;
import java.awt.event.*;


public class vmTerm implements KeyListener, vmConst {
  final static String VERSION = "vmTerm 1.0.1";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  private vmVideo vmOut = null;
  public static final int PC_KEY_TYPED = 0x10000;
  public static final int PC_KEY_PRESSED = 0x20000;
  public static final int PC_KEY_RELEASED = 0x40000;
  public static final int PC_UNICODE_MASK = 0x0FFFF;
  public vmTerm(vmVideo out0) {
    vmOut = out0;
    keybuf = new int[KEYBUFLEN];
    keyg = 0;
    keyp = 0;
    vmOut.addKeyListener(this);
  }
  private final static int KEYBUFLEN = 2048;
  private int[] keybuf;
  private int keyp;
  private int keyg;
  public synchronized int keyEvents() {
    if ((keyp - keyg) >= 0)
      return (keyp - keyg);
    else
      return (keyp + KEYBUFLEN - keyg);
  }

  public synchronized void enqueueKeyEvent(int ekey) {
    if (keyEvents() < KEYBUFLEN - 1) {
      keybuf[keyp] = ekey;
      keyp = (keyp + 1) % KEYBUFLEN;
    }
  }

  public void enqueueKeyTypedEvent(char key) {
    enqueueKeyEvent(PC_KEY_TYPED | ((int) key & PC_UNICODE_MASK));
  }

  public int dequeueKeyEvent() {
    if (keyEvents() > 0) {
      int ekey = keybuf[keyg];

      keyg = (keyg + 1) % KEYBUFLEN;
      return (ekey);
    } else {
      return FALSE;
    }
  }

  public int dequeueKeyTypedEvent() {
    int ekey = dequeueKeyEvent();

    while ((ekey != FALSE) && ((ekey & PC_KEY_TYPED) == 0))
      ekey = dequeueKeyEvent();
    return (ekey);
  }

  public void keyTyped(KeyEvent e) {
    int c = (int) e.getKeyChar();

    if (c < 127) {
      if (c == 13) {
        c = 10;
      }
      enqueueKeyTypedEvent((char) c);
      e.consume();
    }
    return;
  }

  public void keyPressed(KeyEvent e) {
    enqueueKeyEvent(PC_KEY_PRESSED |
      ((int) e.getKeyCode() & PC_UNICODE_MASK));
    e.consume();
  }

  public void keyReleased(KeyEvent e) {
    enqueueKeyEvent(PC_KEY_RELEASED |
      ((int) e.getKeyCode() & PC_UNICODE_MASK));
    e.consume();
  }

  public int readKeyTypedEvent() {
    return dequeueKeyTypedEvent();
  }

  public int readKeyEvent() {
    return dequeueKeyEvent();
  }

  public boolean readyKeyEvents() {
    return (keyEvents() > 0);
  }

  public void close() {
    vmOut.removeKeyListener(this);
  }
}
