/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


public class vmTracer extends Object implements vmConst {
  final static String VERSION = "vmTracer 1.0.48b";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  private vmSystem vmSys;
  private short[] m;
  private short _LAST = 0;
  private short _TRLOW;
  private short _BASE;
  private short _CONTEXIT = -1;
  final static String primName[] = {
      "???", "BYE", "SM/REM", "FM/MOD",
      "_LIT", "EXIT", "EXECUTE", "???",
      "MOVE", "?BRANCH", "BRANCH", "!",
      "@", "C!", "C@", "R>",
      "R@", "RP@", "RP!", ">R",
      "SP@", "SP!", "DROP", "DUP",
      "SWAP", "OVER", "ROT", "0<",
      "AND", "OR", "XOR", "UM+",
      "UM/MOD", "UM*", "1+", "1-",
      "DROP/EXIT", "SWAP/EXIT", "@/EXIT", "+",
      "+!", "PICK", "PUT",
      "SP0!", "RP0!", "SP0@", "RP0@",
      "(SYS)", "NIP",
      "CELL+", "CHAR+", "CELL-", "CHAR-", "CELLS",
      "CHARS", "NOOP",
      "!/EXIT", "+/EXIT",
      "LIT_0", "LIT_1", "LIT_2",
      "ALIGNED", "ALIGN",
      "-", "*",
      "NEGATE", "RSHIFT", "LSHIFT",
      "INVERT",
      "?DUP",
      "2+", "2-",
      "2DUP", "2DROP",
      "ABS", "MIN", "MAX",
      "0=", "=", "<", "U<",
      "2*", "2/",
      "2OVER", "2SWAP", "2@", "2!",
      "S>D",
      "TRACE-STEP",
      "TRACE-ON", "TRACE-OFF", "EXEC-TRACE",
      "SERR!", "TRINI",
    };
  public void initialize(vmSystem sys0,
    short m0[], short last0, short trlow0,
    short base0) {
    vmSys = sys0;
    m = m0;
    _LAST = last0;
    _TRLOW = trlow0;
    _BASE = base0;
    _CONTEXIT = -1;
  }

  private String makeString(int n) {
    StringBuffer sb = new StringBuffer();

    for (int i = 0; i < (m[n] & 0x1F); i++)
      sb.append((char) m[n + 1 + i]);
    String str = new String(sb);

    return str;
  }

  private String getName(int ca, short m[]) {
    int w = m[_LAST];
    String name = "???";

    do {
      if (ca == m[w - 2]) {
        name = makeString(w);
        break;
      } else {
        w = m[w - 1];
      }
    }
    while (w != 0);
    return name;
  }

  private void printName(int inst, short m[]) {
    String name;

    if (inst < 0) {
      name = primName[-inst];
    } else {
      name = getName(inst, m);
    }
    vmSys.vmOut.print(name);
    if (name.length() < 8)
      vmSys.vmOut.print("\t");
  }

  private void printRadix(int val0) {
    switch (_BASE) {
    case 2:
      vmSys.vmOut.print(Integer.toBinaryString(val0 & 0xFFFF));
      break;

    case 8:
      vmSys.vmOut.print(Integer.toOctalString(val0 & 0xFFFF));
      break;

    case 16:
      vmSys.vmOut.print(Integer.toHexString(val0 & 0xFFFF));
      break;

    default:
      vmSys.vmOut.print(Integer.toString(val0));
    }
  }

  public short inform(short inst, short ip, short sp,
    short rp, short SPP, short RPP, short flag) {
    boolean displayLoop = true;
    boolean displaySwitch = true;
    int c = 0;

    if (inst == (-TROFF))
      return flag;
    if (_LAST == 0)
      return flag;
    if (_TRLOW > 0)
      if ((0 < ip) && (ip < m[_TRLOW]))
        return flag;
    if (_CONTEXIT != -1)
      if (_CONTEXIT > rp) {
        flag = TRSTEP;
        _CONTEXIT = -1;
      } else {
        return flag;
      }
    while (displayLoop) {
      if (displaySwitch) {
        vmSys.vmOut.print("\nstack:\t");
        if ((SPP - sp) > 0) {
          printRadix(m[sp]);
        } else {
          vmSys.vmOut.print("---");
        }
        vmSys.vmOut.print("   return:\t");
        if ((rp - RPP) > 0) {
          printRadix(m[rp]);
        } else {
          vmSys.vmOut.print("---");
        }
        vmSys.vmOut.print("   ip:\t");
        printRadix(ip);
        if (inst < 0) {
          vmSys.vmOut.print("\t");
        } else {
          vmSys.vmOut.print("\t: ");
        }
        printName(inst, m);
        switch (-inst) {
        case _LIT:
        case BRAN:
        case QBRAN:
          vmSys.vmOut.print(" ");
          printRadix(m[ip + 1]);
        }
        if ((SPP - sp) > 0) {
          int times = SPP - sp;

          if (times > 4) {
            times = 4;
          }
          for (int i = 1; i < times; i++) {
            vmSys.vmOut.print("\n");
            if (i == 1) {
              vmSys.vmOut.print("(");
              printRadix(SPP - sp);
              vmSys.vmOut.print(")");
            }
            vmSys.vmOut.print("\t");
            printRadix(m[sp + i]);
            if (i == 1) {
              vmSys.vmOut.print("   (");
              printRadix(rp - RPP + 1);
              vmSys.vmOut.print(")");
            } else {
              vmSys.vmOut.print("\t");
            }
            if ((rp - RPP + 1) > i) {
              vmSys.vmOut.print("\t");
              printRadix(m[rp - i]);
              vmSys.vmOut.print("\t\t\t\t");
            } else {
              vmSys.vmOut.print("\t\t\t\t\t\t");
            }
          }
        }
      }
      displaySwitch = true;
      if ((flag & TRSTEP) != 0) {
        c = vmSys.queryKeyPress();
        while (c == TRUE)
          c = vmSys.queryKeyPress();
      } else if ((c = vmSys.queryKeyPress()) == TRUE) {
        c = 10;
      }
      switch (c) {
      case 27:
        flag = NOTRACE;
        displayLoop = false;
        break;

      case 10:
        flag = TRSHOW;
        displayLoop = false;
        break;

      case 32:
        flag = TRSTEP;
        displayLoop = false;
        break;

      case 'n':
      case 'N':
        _CONTEXIT = (short) (rp + 1);
        flag = TRSHOW;
        displayLoop = false;
        break;

      case 'e':
      case 'E':
        _CONTEXIT = rp;
        flag = TRSHOW;
        displayLoop = false;
        break;

      case 'b':
      case 'B':
        _BASE = 2;
        break;

      case 'o':
      case 'O':
        _BASE = 8;
        break;

      case 'd':
      case 'D':
        _BASE = 10;
        break;

      case 'h':
      case 'H':
        _BASE = 16;
        break;

      case '.':
        vmSys.vmOut.print("\n");
        if ((SPP - sp) > 0) {
          for (int i = SPP - 1; i >= sp; i--) {
            vmSys.vmOut.print(" ");
            printRadix(m[i]);
          }
        }
        vmSys.vmOut.print(" <sp");
        displaySwitch = false;
        break;

      case 'r':
      case 'R':
        vmSys.vmOut.print("\n");
        if ((rp - RPP) > 0) {
          for (int i = RPP + 1; i <= rp; i++) {
            vmSys.vmOut.print(" ");
            printRadix(m[i]);
          }
        }
        vmSys.vmOut.print(" <rp");
        displaySwitch = false;
        break;
      }
    }
    return flag;
  }
}
