/*

    myForth v1.0
    Copyright (c) 2002,2003,2004 by Mark A. Washburn.
    All rights reserved.
  
*/



package Washburn.myForth;


import java.awt.*;


public class vmVideo extends Canvas implements vmConst {
  final static String VERSION = "vmVideo 1.0.47a";
  final static String COPYRIGHT =
    "Copyright 2002 Mark A. Washburn All rights reserved.";
  protected int leftPad = 8;
  protected int topPad = 4;
  private int cur = 0;
  private int oldcur = 0;
  private boolean refresh = true;
  private boolean cursorON = true;
  private boolean inScroll = false;
  private final static int scrollDelay = 20;
  protected int rows;
  private int cols;
  private int screen_len, fontsize;
  private Graphics g1;
  private Font f1;
  private int width1;
  protected int height1;
  private int ascent1;
  protected int rowWidth;
  private Graphics[] gbuf;
  private Image[] offscreen;
  private int drawEnabled = 0;
  private Image g1img;
  private Graphics g1buf;
  protected Dimension g1dim;
  private int scrollOff;
  private char[] cGenX;
  private boolean cGenEnabled = true;
  private Graphics[] cBufX;
  private Image[] cImageX;
  private int cGenXFGRGB = -1;
  private int cGenXBGRGB = -1;
  private int cGenLast = 0;
  final static int C_GEN_LAST_MAXIMUM = 2;
  private Graphics[][] cBuf1;
  private Image[][] cImage1;
  private int cGen1FGRGB[];
  private int cGen1BGRGB[];
  private Color FGColor;
  private Color BGColor;
  private Color cursorColor;
  private int logEnabled = 0;
  public vmVideo(int rows, int cols, int fontsize, int initFG, int initBG) {
    super();
    this.rows = rows;
    this.cols = cols;
    this.fontsize = fontsize;
    FGColor = new Color(initFG);
    BGColor = new Color(initBG);
    setForeground(FGColor);
    setBackground(BGColor);
    f1 = new Font("Monospaced", Font.PLAIN, fontsize);
    if (!f1.getName().equals("Monospaced"))
      f1 = new Font("Courier", Font.PLAIN, fontsize);
    FontMetrics fm1 = this.getFontMetrics(f1);

    width1 = fm1.charWidth('X');
    height1 = fm1.getHeight();
    ascent1 = fm1.getAscent();
    rowWidth = cols * width1;
    screen_len = rows * cols;
    setSize((leftPad * 2) + (cols * width1),
      (topPad * 2) + (rows * height1));
  }

  private void cGen() {
    int i, z, k;

    for (i = 0; i < C_GEN_LAST_MAXIMUM; i++) {
      if (cGen1FGRGB[i] == cGenXFGRGB && cGen1BGRGB[i] == cGenXBGRGB) {
        cImageX = cImage1[i];
        cBufX = cBuf1[i];
        break;
      }
    }
    if (i == C_GEN_LAST_MAXIMUM) {
      cGenLast++;
      if (cGenLast > C_GEN_LAST_MAXIMUM)
        cGenLast = 1;
      k = cGenLast - 1;
      for (i = 32; i < 128; i++) {
        z = i - 32;
        cImageX[z].flush();
        cBuf1[k][z].setColor(BGColor);
        cBuf1[k][z].fillRect(0, 0, width1, height1);
        cBuf1[k][z].setColor(FGColor);
        cBuf1[k][z].drawChars(cGenX, z, 1, 0, ascent1);
      }
      cImageX = cImage1[k];
      cBufX = cBuf1[k];
      cGen1FGRGB[k] = cGenXFGRGB;
      cGen1BGRGB[k] = cGenXBGRGB;
    }
    cGenEnabled = false;
  }

  private void drawCursor() {
    if (refresh && !inScroll) {
      int x = leftPad + (cur % cols) * width1;
      int y = topPad + ((cur / cols) * height1) + ascent1;

      g1.setColor(cursorColor);
      g1.drawLine(x, y, x + width1 - 1, y);
    }
  }

  private void eraseCursor() {
    if (refresh) {
      int currow = (cur / cols);
      int x = (cur % cols) * width1;
      int y = (currow * height1);
      int x2 = leftPad + x;
      int y2 = topPad + y;

      if (drawEnabled == 0) {
        int currow2 = currow + scrollOff;

        if (currow2 > rows - 1)
          currow2 = currow2 - rows;
        g1.drawImage(offscreen[currow2], x2, y2,
          x2 + width1, y2 + height1,
          x, 0, x + width1, height1, null);
      } else {
        g1.drawImage(g1img, x2, y2, x2 + width1, y2 + height1,
          x2, y2, x2 + width1, y2 + height1, null);
      }
    }
  }

  protected void setCursorEnable(boolean flag) {
    if (flag) {
      cursorON = true;
      cursorColor = new Color(BGColor.getRGB() ^ 0x00FFFFFF);
      drawCursor();
    } else {
      eraseCursor();
      cursorON = false;
    }
  }

  private void setCur(int x, int y) {
    eraseCursor();
    cur = (y * cols) + x;
    oldcur = cur;
    drawCursor();
  }

  protected void setRefresh(boolean flag) {
    if (flag != refresh) {
      refresh = flag;
      if (refresh) {
        setForeground(FGColor);
        setBackground(BGColor);
        repaint();
      }
    }
  }

  protected void clearPage() {
    int i;

    eraseCursor();
    cur = 0;
    setForeground(FGColor);
    setBackground(BGColor);
    setFG(getFG());
    setBG(getBG());
    drawCursor();
    oldcur = 0;
    if (drawEnabled == 0) {
      for (i = 0; i < rows; i++) {
        gbuf[i].setColor(BGColor);
        gbuf[i].fillRect(0, 0, rowWidth, height1);
      }
      scrollOff = 0;
    } else {
      g1buf.setColor(BGColor);
      g1buf.fillRect(0, 0, g1dim.width, g1dim.height);
    }
    if (refresh)
      paintScreen();
  }

  private void scroll() {
    int i;
    int x, y;

    inScroll = true;
    if (drawEnabled == 0) {
      if (scrollOff + 1 < rows) {
        gbuf[scrollOff].setColor(BGColor);
        gbuf[scrollOff].fillRect(0, 0, rowWidth, height1);
        scrollOff++;
      } else {
        gbuf[scrollOff].setColor(BGColor);
        gbuf[scrollOff].fillRect(0, 0, rowWidth, height1);
        scrollOff = 0;
      }
    } else {
      x = leftPad;
      y = topPad + height1;
      g1buf.copyArea(x, y, rowWidth, height1 * (rows - 1), 0, -height1);
      y = topPad + ((rows - 1) * height1);
      g1buf.setColor(BGColor);
      g1buf.fillRect(x, y, rowWidth, height1);
    }
    cur = screen_len - cols;
    oldcur = cur - 1;
  }

  protected void setFG(int color0) {
    FGColor = new Color(color0);
    cGenXFGRGB = FGColor.getRGB();
    cGenEnabled = true;
  }

  protected void setBG(int color0) {
    BGColor = new Color(color0);
    cGenXBGRGB = BGColor.getRGB();
    cGenEnabled = true;
  }

  protected int getFG() {
    return FGColor.getRGB() & 0xFFFFFF;
  }

  protected int getBG() {
    return BGColor.getRGB() & 0xFFFFFF;
  }

  protected void setLogEnable(int f0) {
    logEnabled = f0;
  }

  protected int getRefresh() {
    if (refresh) {
      return -1;
    } else {
      return 0;
    }
  }

  protected void atXY(int x0, int y0) {
    setCur(x0, y0);
    paintScreen();
  }

  protected void setRefresh(int f0) {
    if (f0 != 0) {
      if (!refresh) {
        refresh = true;
        paintScreen();
      }
    } else
      refresh = false;
  }

  protected void setDrawEnable(int f0) {
    if (f0 != drawEnabled) {
      drawEnabled = f0;
      paintScreen();
    }
  }

  protected void drawLine(int x1, int y1, int x2, int y2) {
    if (refresh) {
      synchronized (g1) {
        g1.setColor(FGColor);
        g1.drawLine(x1, y1, x2, y2);
      }
    }
    g1buf.setColor(FGColor);
    g1buf.drawLine(x1, y1, x2, y2);
  }
  StringBuffer escS0 = null;
  String escS1 = "";
  int escLevel = 0;
  private void doEscape(int c) {
    int val0;
    int val1;
    int currow;

    if (c == 27) {
      if (escLevel == 1) {
        escLevel = 0;
      } else {
        escLevel = 1;
        escS1 = "";
      }
      escS0 = null;
      return;
    }
    if (escLevel == 1) {
      if (c == '[') {
        escLevel = 2;
      } else {
        escS0 = null;
        escLevel = 0;
      }
      return;
    }
    if (c == ';') {
      escLevel = 3;
      if (escS0 != null) {
        escS1 = new String(escS0);
      } else {
        escLevel = 0;
      }
      escS0 = null;
      return;
    }
    if (c >= '0' && c <= '9') {
      if (escS0 == null)
        escS0 = new StringBuffer();
      escS0.append((char) c);
      return;
    }
    try {
      switch (c) {
      case 'H':
      case 'f':
        if ((escLevel == 2) && (escS0 == null)) {
          setCur(0, 0);
        } else if (escLevel == 3) {
          if ((escS0 != null) && (escS1.length() > 0)) {
            val1 = Integer.parseInt(escS1, 10);
            val0 = Integer.parseInt(new String(escS0), 10);
            setCur(val0, val1);
          }
        }
        break;

      case 'A':
        if ((escLevel == 2) && (escS0 != null)) {
          val0 = Integer.parseInt(new String(escS0), 10);
          eraseCursor();
          while (val0-- > 0) {
            if (cur / cols > 0) {
              cur = cur - cols;
            }
          }
          drawCursor();
        }
        break;

      case 'B':
        if ((escLevel == 2) && (escS0 != null)) {
          val0 = Integer.parseInt(new String(escS0), 10);
          eraseCursor();
          while (val0-- > 0) {
            if (cur / cols < rows) {
              cur = cur + cols;
            }
          }
          drawCursor();
        }
        break;

      case 'C':
        if ((escLevel == 2) && (escS0 != null)) {
          val0 = Integer.parseInt(new String(escS0), 10);
          eraseCursor();
          currow = (cur / cols);
          while (val0-- > 0) {
            if ((cur + 1) / cols == currow) {
              cur++;
            }
          }
          drawCursor();
        }
        break;

      case 'D':
        if ((escLevel == 2) && (escS0 != null)) {
          val0 = Integer.parseInt(new String(escS0), 10);
          eraseCursor();
          currow = (cur / cols);
          while (val0-- > 0) {
            if ((cur - 1) / cols == currow) {
              cur--;
            }
          }
          drawCursor();
        }
        break;

      case 'J':
        if ((escLevel == 2) && (escS0 != null)) {
          val0 = Integer.parseInt(new String(escS0), 10);
          if (val0 == 2)
            clearPage();
        }
        break;

      case 'K':
        if ((escLevel == 2) && (escS0 == null)) {
          eraseCursor();
          currow = (cur / cols);
          val0 = cur;
          escLevel = 0;
          while ((cur / cols) == currow) {
            write(' ');
          }
          cur = val0;
          drawCursor();
        }
        break;
      }
    } catch (Exception e) {}
    escS0 = null;
    escLevel = 0;
    return;
  }

  public void write(int c) {
    int x, y, x2, y2;
    int oldrow;
    int oldrow2;

    oldcur = cur;
    if (c == 27 || escLevel != 0) {
      doEscape(c);
      return;
    }
    if ((c < 32) || (c > 127)) {
      eraseCursor();
      if (c == '\t') {
        cur = (cur + 8) & 0xFFF8;
      } else if (c == 10) {
        cur = (cur / cols) * cols + cols;
      } else if (c == 8) {
        if (cur > 0)
          cur = cur - 1;
      }
      if (cur >= screen_len) {
        scroll();
        if (refresh)
          repaint(scrollDelay);
      }
      drawCursor();
      if (logEnabled != 0 && c == 10)
        System.err.print("\n");
      return;
    }
    if (logEnabled != 0) {
      System.err.print((char) c);
      if (((cur + 1) % cols) == 0)
        System.err.print("\n");
    }
    if (cGenEnabled)
      cGen();
    if (drawEnabled == 0) {
      oldrow = (oldcur / cols);
      oldrow2 = oldrow + scrollOff;
      if (oldrow2 > rows - 1)
        oldrow2 = oldrow2 - rows;
      x = (oldcur % cols) * width1;
      y = (oldrow * height1);
      x2 = leftPad + x;
      y2 = topPad + y;
      if (refresh && !inScroll)
        synchronized (g1) {
          g1.drawImage(cImageX[c - 32], x2, y2, null);
        }
      gbuf[oldrow2].drawImage(cImageX[c - 32], x, 0, null);
    } else {
      oldrow = (oldcur / cols);
      x = (oldcur % cols) * width1;
      y = (oldrow * height1);
      x2 = leftPad + x;
      y2 = topPad + y;
      if (refresh && !inScroll)
        synchronized (g1) {
          g1.drawImage(cImageX[c - 32], x2, y2, null);
        }
      g1buf.drawImage(cImageX[c - 32], x2, y2, null);
    }
    cur++;
    if (cur >= screen_len) {
      scroll();
      if (refresh)
        repaint(scrollDelay);
    }
    if (refresh && cursorON) {
      drawCursor();
    }
  }

  protected void print(String s) {
    int i, len = s.length();

    for (i = 0; i < len; i++)
      write(s.charAt(i));
  }

  public void update(Graphics g) {
    paint(g);
  }

  public void paint(Graphics g) {
    if (refresh)
      paintScreen();
  }

  protected void paintScreen() {
    int row;
    int x, y;

    if (!refresh)
      return;
    if (drawEnabled == 0) {
      if (scrollOff == 0) {
        for (row = 0; row < rows; row++) {
          y = topPad + (row * height1);
          g1.drawImage(offscreen[row], leftPad, y, null);
        }
      } else {
        for (row = scrollOff; row < rows; row++) {
          y = topPad + ((row - scrollOff) * height1);
          g1.drawImage(offscreen[row], leftPad, y, null);
        }
        for (row = 0; row < scrollOff; row++) {
          y = topPad + ((rows - scrollOff + row) * height1);
          g1.drawImage(offscreen[row], leftPad, y, null);
        }
      }
    } else {
      g1.drawImage(g1img, 0, 0, null);
    }
    if (cursorON) {
      if (cur < screen_len) {
        x = leftPad + (cur % cols) * width1;
        y = topPad + ((cur / cols) * height1) + ascent1;
        g1.setColor(cursorColor);
        g1.drawLine(x, y, x + width1 - 1, y);
      }
    }
    inScroll = false;
  }

  public void addNotify() {
    synchronized (getTreeLock()) {
      super.addNotify();
    }
    int i;

    g1 = this.getGraphics();
    g1dim = this.getSize();
    g1img = this.createImage(g1dim.width, g1dim.height);
    g1buf = g1img.getGraphics();
    g1buf.setFont(f1);
    g1buf.setColor(BGColor);
    g1buf.fillRect(0, 0, g1dim.width, g1dim.height);
    offscreen = new Image[rows];
    gbuf = new Graphics[rows];
    for (i = 0; i < rows; i++) {
      offscreen[i] = this.createImage(rowWidth, height1);
      gbuf[i] = offscreen[i].getGraphics();
      gbuf[i].setFont(f1);
      gbuf[i].setColor(BGColor);
      gbuf[i].fillRect(0, 0, rowWidth, height1);
    }
    cGenX = new char[128 - 32 + 1];
    cImage1 = new Image[C_GEN_LAST_MAXIMUM][128 - 32 + 1];
    cBuf1 = new Graphics[C_GEN_LAST_MAXIMUM][128 - 32 + 1];
    cGen1FGRGB = new int[C_GEN_LAST_MAXIMUM];
    cGen1BGRGB = new int[C_GEN_LAST_MAXIMUM];
    for (i = 32; i < 128; i++)
      cGenX[i - 32] = (char) i;
    for (i = 0; i < C_GEN_LAST_MAXIMUM; i++) {
      cGen1FGRGB[i] = -1;
      cGen1BGRGB[i] = -1;
      for (int i2 = 32; i2 < 128; i2++) {
        int z = i2 - 32;

        cImage1[i][z] = this.createImage(width1, height1);
        cBuf1[i][z] = cImage1[i][z].getGraphics();
        cBuf1[i][z].setFont(f1);
      }
    }
    cGenEnabled = true;
    cGen();
  }

  public void removeNotify() {
    synchronized (getTreeLock()) {
      int i;

      for (i = 0; i < C_GEN_LAST_MAXIMUM; i++) {
        for (int i2 = 0; i2 < (128 - 32); i2++) {
          cBuf1[i][i2].dispose();
          cImage1[i][i2].flush();
        }
        cBufX[i].dispose();
        cImageX[i].flush();
      }
      for (i = 0; i < rows; i++) {
        gbuf[i].dispose();
        offscreen[i].flush();
      }
      g1buf.dispose();
      g1img.flush();
      g1.dispose();
      super.removeNotify();
    }
  }
}
